import numpy as np
from periodictable import cromermann
from grsq.accel import numbye

class Debye:
    ''' Discrete Interatomic-Distance Debye Scattering. 

    .. code-block:: none   

            ___N  ___N
            \    \                       sin(q · r_ij)
      I(q) = )    )   f_i(q) · f_j(q) · ---------------
            /__  /__                       q · r_ij
            i=1  j=1

    where:
       I(q)    : Scattering intensity as a function of the momentum transfer q.
       N       : Total number of atoms.
       f_i, f_j: Atomic form actors for atoms i and j, respectively.
       r_ij    : Distance between atoms i and j.
       q       : Momentum transfer, or scattering vector.

       The equation involves a double summation over all atoms i and j in the structure.
       The terms in the equation represent the wave nature of the scattered intensity,

       Usage:

       .. code-block:: python   
       
           >>> from grsq.debye import Debye
           >>> from ase.io import read
           >>> atoms = read('./tests/data/xray_2particle/test.xyz')  # read e.g. xyz file as ASE atoms object
           >>> deb = Debye(qvec=np.arange(0, 20, 0.01))
           >>> i_deb = deb.debye(atoms)
    '''

    def __init__(self, qvec=None, custom_cm=False):
        if qvec is None:
            qvec = np.arange(0.0001, 10, 0.01)
        self.qvec = qvec  # q vector
        self.f0 = None
        self.f0_elements = None
        self.custom_cm = custom_cm

        assert len(qvec.shape) == 1, 'Q-vector shape must be (N, )'

        # Cromer mann coefficients - look in f0_cromermann.txt
        # https://www.classe.cornell.edu/~dms79/x-rays/f0_CromerMann.txt
        # only used if custom_cm = True, otherwise the cromermann method
        # from periodictable will take care of it
        self.cm = {
            'Pt': {'a': np.array([27.00590, 17.76390, 15.71310, 5.783700]),
                   'b': np.array([1.512930, 8.811740, 0.4245930, 38.61030]),
                   'c': 11.68830},
            'Ag': {'a': np.array([19.28080, 16.68850, 4.804500, 1.046300]),
                   'b': np.array([0.6446000, 7.472600, 24.66050, 99.81560]),
                   'c': 5.179000},
            'Ir': {'a': np.array([30.70580, 15.55120, 14.23260, 5.536720]),
                   'b': np.array([1.309230, 6.719830, 0.1672520, 17.49110]),
                   'c': 6.968240},  # TODO: This is Ir4+.
            'Fe': {'a': np.array([11.04240, 7.374000, 4.134600, 0.4399000]),
                   'b': np.array([4.653800, 0.3053000, 12.05460, 31.28090]),
                   'c': 1.009700},  # TODO: This is Fe2+. Needs flexibility
            'P':  {'a': np.array([6.434500, 4.179100, 1.780000, 1.490800]),
                   'b': np.array([1.906700, 27.15700, 0.5260000, 68.16450]),
                   'c': 1.114900},
            'O':  {'a': np.array([3.048500, 2.286800, 1.546300, 0.8670000]),
                   'b': np.array([13.27710, 5.701100, 0.3239000, 32.90890]),
                   'c': 0.2508000},
            'C':  {'a': np.array([2.310000, 1.020000, 1.588600, 0.8650000]),
                   'b': np.array([20.84390, 10.20750, 0.5687000, 51.65120]),
                   'c': 0.2156000},
            'N':  {'a': np.array([12.21260, 3.13220, 2.01250, 1.166300]),
                   'b': np.array([5.701e-3, 9.89330, 28.9975, 0.582600]),
                   'c': -11.52900},
            'H':  {'a': np.array([0.4930020, 0.3229120, 0.1401910, 4.081e-2]),
                   'b': np.array([10.51090, 26.12570, 3.142360, 57.79970]),
                   'c': 3.0380001e-03},
            'Na': {'a': np.array([3.256500, 3.936200, 1.399800, 1.003200]),
                   'b': np.array([2.667100, 6.115300, 0.2001000, 14.03900]),
                   'c': 0.4040000},
            'I':  {'a': np.array([20.23320, 18.99700, 7.806900, 2.886800]),  # I-
                   'b': np.array([4.347000, 0.3814000, 27.76600, 66.87760]),
                   'c': 4.071200},
            'I0': {'a': np.array([20.14720, 18.99490, 7.513800, 2.273500]),
                   'b': np.array([4.347000, 0.3814000, 27.76600, 66.87760]),
                   'c': 4.071200},
            'Tl': {'a': np.array([21.39850, 20.47230, 18.74780, 6.828470]),   # Tl+
                   'b': np.array([1.471100, 0.5173940, 7.434630, 28.84820]),
                   'c': 12.52580},
            'Si': {'a': np.array([4.439180, 3.203450, 1.194530, 0.4165300]),   # Si4+
                   'b': np.array([1.641670, 3.437570, 0.2149000, 6.653650]),
                   'c': 0.7462970},
            'B':  {'a': np.array([2.054500, 1.332600, 1.097900, 0.7068000]),
                   'b': np.array([23.21850, 1.021000, 60.34980, 0.1403000]),
                   'c': -0.1932000},
            'S':  {'a': np.array([6.905300, 5.203400, 1.437900, 1.58630]),
                   'b': np.array([1.467900, 22.21510, 0.2536000, 56.17200]),
                   'c': 0.8669000},
            'Cl': {'a': np.array([11.46040, 7.196400, 6.255600, 1.645500]),
                   'b': np.array([1.0400000E-02, 1.166200, 18.51940, 47.77840]),
                   'c': -9.557400},
            'Ar': {'a': np.array([7.484500, 6.772300, 0.6539000, 1.644200]),
                   'b': np.array([0.9072000, 14.84070, 43.89830, 33.39290]),
                   'c': 1.444500}
                   }

    def atomic_f0(self, atom):
        ''' Atomic form factors '''
        if self.custom_cm:
            f0 = self.cm[atom]['c']
            a = self.cm[atom]['a']
            b = - self.cm[atom]['b']

            q2 = (self.qvec / (4 * np.pi))**2
            for n in range(4):
                f0 += a[n] * np.exp(b[n] * q2)
        else:
            f0 = cromermann.fxrayatq(atom, self.qvec)

        return f0

    def update_f0(self, atoms, custom_elements=None):
        ''' Update atomic form factors in the object '''
        if self.f0 is None:
            self.f0_elements = atoms.get_chemical_symbols()
            if custom_elements:
                self.f0_elements = custom_elements
            f0 = np.array([self.atomic_f0(sym) for sym in self.f0_elements])
            self.f0 = f0
        else:  # if the atoms have changed:
            these_syms = atoms.get_chemical_symbols()
            if these_syms != self.f0_elements:
                self.f0_elements = these_syms
                f0 = np.array([self.atomic_f0(sym) for sym in self.f0_elements])
                self.f0 = f0

    def debye_numba(self, atoms, custom_elements=None):
        ''' Faster implementation, using numba. 
            Useful if you need to do a _lot_ of evaluations. 
        '''

        self.update_f0(atoms, custom_elements)
        f0 = self.f0

        rs = atoms.get_positions()
        s = numbye(self.qvec, len(atoms), f0, rs, np.zeros(len(self.qvec)))
        return s

    def debye(self, atoms, custom_elements=None):
        ''' Debye Scattering from ASE atoms object'''

        s = np.zeros(len(self.qvec))
        self.update_f0(atoms, custom_elements)
        f0 = self.f0

        rs = atoms.get_all_distances()
        for q, qq in enumerate(self.qvec):
            for i in range(len(atoms)):
                s[q] += np.sum(f0[i, q] * f0[:, q] *
                               np.sinc(qq * rs[i, :] / np.pi))

        return s

    def debye_selective(self, atoms, idx1, idx2, custom_elements=None):
        ''' Only terms between atoms in lists idx1 and idx2
            (but not between atoms in either list) '''

        s = np.zeros(len(self.qvec))
        self.update_f0(atoms, custom_elements)
        f0 = self.f0

        rs = atoms.get_all_distances()
        for q, qq in enumerate(self.qvec):
            for i in idx1:
                s[q] += np.sum(f0[i, q] * f0[idx2, q] *
                               np.sinc(qq * rs[i, idx2] / np.pi))

        return s




import numpy as np 

class Damping:
    '''
    Class that applies a damping window function to ( g(r) - 1 ) in the
    generalized debye integral. 

    Instantiate the object and add it to your RDF(), 
    e.g. 

    .. code-block:: python 

        >>> L = 25  # for a 50 Å cell
        >>> damp = Damping('lorch', L=L)
        >>> rdf = RDF(..., damp=damp)

        
    You can also instruct an ``RDFSet`` to use the same damping 
    for each RDF to calculate the full scattering signal 

    .. code-block:: python   
    
        >>> rdfs = RDFSet(...)
        >>> damp = Damping('lorch', L=25)
        >>> rdfs.get_iq(damping=damp)

    style: str
        | can be either of these:
                | 'lorch': sin(pi*L) / pi*L
                | 'dhabal': as in 10.1039/c6cp07599a
                | 'zederkof': as above but with an r_cut as well. 10.1021/jacs.2c04505
                | 'panman': Exponential fade function, e.g. PhysRevLett.125.226001

    L: Float (for Lorch damping)
        | Damping factor for the lorch damping, usually 1/2 of
        | the box length.
    r_max: Float  (for Dhabal, Zederkof, Panman)
        | The r-value where the rdf should be 0
    r_cut: Float (only for zederkof damping)
        | When to start the smooth cutoff. Everything before
        | is untouched.

    '''
    def __init__(self, style, L=None, r_max=None, r_cut=None):
        self.style = style
        self.L = L
        self.r_max = r_max
        self.r_cut = r_cut

    def lorch(self, R):
        ''' Lorch-like damping '''
        damp = self.L
        return np.sin(np.pi * R / damp) / (np.pi * R / damp)

    def panman(self, R):
        ''' Damping like in PhysRevLett.125.226001 '''
        fade = self.L
        mask = R > fade
        damping = np.ones(len(R))
        damping[mask] = np.exp(-(R[mask] / fade - 1)**2)
        return damping

    def dhabal(self, R):
        ''' Damping like in 10.1039/c6cp07599a '''
        r_max = self.r_max
        fac = (1 - 3*(R/r_max)**2) * (R <= r_max/3) +\
               3 / 2 * (1-2*(R/r_max) +\
               (R/r_max)**2) * ((r_max / 3 < R) & (r_max > R))

        return fac

    def zederkof(self, R):
        ''' Dampingl ike in 10.1021/jacs.2c04505 '''
        r_max = self.r_max
        r_cut = self.r_cut

        fac = 1 * (R < r_cut) +\
              (1 - 3 * ((R - r_cut) / (r_max - r_cut))**2) * ((R <= 2 * r_cut / 3 + r_max / 3) & (R >= r_cut)) +\
              3 / 2 * (1 - 2 * ((R - r_cut) / (r_max - r_cut)) +
              (( R - r_cut) / (r_max - r_cut))**2) * ((R > 2 * r_cut / 3 + r_max / 3) & (r_max > R))
        return fac

    def damp(self, R):
        ''' Apply the chosen damping '''
        if self.style == 'lorch':
            damping = self.lorch
        elif self.style == 'dhabal':
            damping = self.dhabal
        elif self.style == 'zederkof':
            damping = self.zederkof
        elif self.style == 'panman':
            damping = self.panman
        else:
            raise RuntimeError(f'Damping style not understood: {self.style}')

        return damping(R)



"""
Unit and regression test for the grsq package.

run:
    `python -m pytest --cov-report=html --cov=src .`
in the root dir of the package to generate covereage report
"""

# Import package, test suite, and other packages as needed
import sys, copy, glob
import pytest
import numpy as np
from ase.io import read

try: # this is NOT how this is supposed to be done I think...
    from grsq import Debye, Damping
    from grsq import RDF, RDFSet, rdfset_from_dir
except:
    sys.path.append('src/')
    from grsq import Debye
    from grsq import Debye, Damping
    from grsq import RDF, RDFSet, rdfset_from_dir


def test_grsq_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "grsq" in sys.modules


def test_twoparticle_solute():
    atoms = read('tests/data/xray_2particle/test.xyz')
    deb = Debye(qvec=np.arange(0, 20, 0.01))
    i_deb = deb.debye(atoms)

    stoich = {'Pt_u':2}
    V = 10**3

    rdf_dat = np.genfromtxt('tests/data/xray_2particle/gPt_u-Pt_u.dat')

    rdf = RDF(rdf_dat[:, 0], rdf_dat[:, 1], 'Pt', 'Pt', 'solute', 'solute', qvec=deb.qvec)
    rdf.n1 = 2
    rdf.n2 = 2
    rdf.volume = V

    iq = rdf.i_term1()
    iq += rdf.i_term2()
    assert np.max(np.abs(deb.qvec, i_deb - iq)) < 20


def test_threeparticle_solute():
    ''' Tests the loader function and the RDFset Class '''
    qvec = np.arange(0, 20, 0.01)

    rdfs = rdfset_from_dir('tests/data/xray_3particle/',
                           'tests/data/xray_3particle/ptptar.xyz', volume=10**3)

    atoms = read('tests/data/xray_3particle/ptptar.xyz')
    deb = Debye(qvec=qvec)
    i_deb = deb.debye(atoms)
    num_electrons_squared = sum(atoms.get_atomic_numbers())**2

    i_g_u = rdfs.get_solute(qvec)  # solute-only (of solute only RDFs)
    i_g = rdfs.get_iq(qvec)

    assert np.sum(np.abs(i_g_u - i_g)) < 1e-9
    assert np.max(np.abs(deb.qvec, i_deb - i_g)) < 20
    assert np.abs(i_deb[0] - num_electrons_squared) < 25
    assert np.abs(i_g[0] - num_electrons_squared) < 25

def test_coordnum():
    ''' Test that the coordination number method gets counts 2 Pt from Ar 
        in the PtPtAr system
    '''

    rdfs = rdfset_from_dir('tests/data/xray_3particle/',
                           'tests/data/xray_3particle/ptptar.xyz', volume=10**3)

    rdf = rdfs[0]
    assert rdf.coordination_number(0, 100) - 2 < 1e-6

def test_water():
    ''' Testing versus reference values saved in npy '''
    ig_v_correct, ig_a_correct, igc_v_correct = np.load('tests/data/xray_water/watertest.npy')
    V = 122900.85207774633
    stoich = {'H_v': 8190, 'O_v': 4095}
    qvec = np.arange(0, 10, 0.05)

    rdfs = rdfset_from_dir('tests/data/xray_water/', volume=V, stoich=stoich)
    ig_v = rdfs.get_solvent(qvec)
    ig_a = rdfs.get_iq(qvec)

    # Test vdv correction against reference
    rdfs.vdv_correct()
    igc_v = rdfs.get_solvent(qvec)

    assert np.sum(ig_v_correct - ig_v) < 1e-9
    assert np.sum(ig_a_correct - ig_a) < 1e-9
    assert np.sum(igc_v_correct - igc_v) < 1e-9

def test_adding():
    ''' Testing addition (and slicing) '''
    ig_v_correct, ig_a_correct, igc_v_correct = np.load('tests/data/xray_water/watertest.npy')
    V = 122900.85207774633
    stoich = {'H_v': 8190, 'O_v': 4095}
    qvec = np.arange(0, 10, 0.05)

    rdfs = rdfset_from_dir('tests/data/xray_water/', volume=V, stoich=stoich)
    rdfs_new = rdfs[:2] + rdfs[2:]  # slice and add

    old = rdfs.get_iq(qvec)
    new = rdfs_new.get_iq(qvec)

    assert ((old - new ) == 0).all()

def test_corrections_dampings():
    ''' Test that all volume corrections and FT windows
        gives results within 12 % of each other when tuned
        correctly.  On MDAnalysis sampled water. '''
    V = 122900.85207774633
    stoich = {'H_v': 8190, 'O_v': 4095}
    rdfs_raw = rdfset_from_dir('tests/data/mda_water/', volume=V, stoich=stoich)
    r_max = rdfs_raw[0].r[-1]

    lrh = Damping('lorch', L=r_max)
    zkf = Damping('zederkof', r_max=r_max, r_cut=r_max-1)
    dhb = Damping('dhabal', r_max=r_max)
    fad = Damping('panman', L=10)
    dampings = [lrh, zkf, dhb, fad]
    corrections = ('volume', 'perera', 'vegt')

    qvec = np.arange(0, 6, 0.01)
    results = np.zeros((len(qvec), 13))

    ct = 0
    for c, corr in enumerate(corrections):
        for d, damp in enumerate(dampings):
            rdfs = copy.deepcopy(rdfs_raw)
            for rdf in rdfs.values():
                rdf.rho_norm = 'N'
                r_first_nonzero = rdf.r[np.where(rdf.g > 0)[0]][0]
                r_c = r_first_nonzero - 0.5
                corr_vars = ({'Ri':r_c}, {'Ri':r_first_nonzero, 'R_avg':10.6579}, {})
                rdf.g = rdf.correct(method=corr, **(corr_vars[c]))
                rdf.damp = damp

            iq = rdfs.get_solvent(qvec=qvec)
            results[:, ct] = iq
            ct += 1

    rdfs = copy.deepcopy(rdfs_raw)
    for rdf in rdfs.values():
        rdf.rho_norm = 'N'
        r_first_nonzero = rdf.r[np.where(rdf.g > 0)[0]][0]
        r_c = r_first_nonzero - 0.5
        opt = rdf.fit(r_c, fit_start=15, fit_stop=r_max)
        print(opt.x[0], r_c)
        rdf.g = rdf.correct(method='volume', Ri=opt.x[0])
        rdf.damp = lrh
    iq = rdfs.get_solvent(qvec=qvec)
    results[:, ct] = iq

    assert np.max(np.abs(np.diff(results, axis=-1) / np.mean(results, axis=-1)[:, None])) < 0.12


def test_cross_dv():
    ''' Assert that the cross term and DV term of an LJ fluid
        are the same as at the point of publication
    '''
    def get_data(fdir):
        uv_files = sorted(glob.glob(fdir + '/*gAr_u-Ar_v.dat'))
        vv_files = sorted(glob.glob(fdir + '/*gAr_v-Ar_v.dat'))
        out = []
        for files in (uv_files, vv_files):
            tmp = np.genfromtxt(files[0])
            rdfs = np.zeros((tmp.shape[0], tmp.shape[1], len(files)))
            for f, fil in enumerate(files):
                dat = np.genfromtxt(fil)
                rdfs[:, :, f] = dat
            out.append(np.mean(rdfs, axis=-1))
        return out

    qvec = np.arange(0, 6, 0.005)
    dat = get_data('tests/data/xray_lj/')
    V = 100**3
    rdf_uv = RDF(r=dat[0][:, 0], g=dat[0][:, 1],
                    name1='Ar', name2='Ar', region1='solute', region2='solvent',
                    n1=1, n2=20712, volume=V, qvec=qvec)

    rdf_vv = RDF(r=dat[1][:, 0], g=dat[1][:, 1],
                    name1='Ar', name2='Ar', region1='solvent', region2='solvent',
                    n1=20712, n2=20712, volume=V, qvec=qvec)

    rdfs = RDFSet()

    rdfs.add_rdf(rdf_uv)
    rdfs.add_flipped(rdf_uv)
    rdfs.add_rdf(rdf_vv)
    rdfs.vdv_correct()

    cross_clc = rdfs.get_cross(qvec=qvec)
    dispv_clc = rdfs.get_dv(qvec=qvec)

    cross_ref = np.load('tests/data/xray_lj/cross.npy')
    dispv_ref = np.load('tests/data/xray_lj/dv.npy')

    assert (np.abs(cross_clc - cross_ref) < 1e-10).all()
    assert (np.abs(dispv_clc - dispv_ref) < 1e-10).all()


def test_classes():
    ''' Test base funcionality '''
    V = 122900.85207774633
    stoich = {'H_v': 8190, 'O_v': 4095}

    rdfs = rdfset_from_dir('tests/data/mda_water/', volume=V, stoich=stoich)
    rdfs.show()
    rdfs[0].get_info()
    assert len(rdfs) == 4
    print(rdfs[0])
    repr(rdfs[0])
    print(rdfs[0])
    print(rdfs[:2])
    print(rdfs[2:-1])
    print(rdfs[2:])
    print(rdfs[0::2])


def test_builtins():
    ''' Testing setitem and eq '''
    V = 122900.85207774633
    stoich = {'H_v': 8190, 'O_v': 4095}

    rdfs_1 = rdfset_from_dir('tests/data/mda_water/', volume=V, stoich=stoich)
    rdfs_2 = RDFSet()
    for key, rdf in rdfs_1.items():
        rdfs_2[key] = copy.deepcopy(rdf) 
    assert rdfs_1 == rdfs_2

    rdfs_2 = copy.deepcopy(rdfs_1)
    rdfs_2[0].n1 = 9999
    assert not (rdfs_1 == rdfs_2)

    rdfs_2 = copy.deepcopy(rdfs_1)
    rdfs_2[0].name1 = 'XX'
    assert not (rdfs_1 == rdfs_2 )

    rdfs_2 = copy.deepcopy(rdfs_1)
    rdfs_2[0].g = np.zeros(3)
    assert not (rdfs_1 == rdfs_2)

    rdfs_2 = copy.deepcopy(rdfs_1)
    rdfs_2[0].r = np.zeros(3)
    assert not (rdfs_1 == rdfs_2)

    rdfs_2 = copy.deepcopy(rdfs_1)
    rdfs_2[0].r *= 0 
    assert not (rdfs_1 == rdfs_2) 

    rdfs_2 = copy.deepcopy(rdfs_1)
    rdfs_2[0].g *= 0 
    assert not (rdfs_1 == rdfs_2) 

    rdfs_2 = RDFSet()
    for key, rdf in rdfs_1[:1].items():
        rdfs_2[key] = copy.deepcopy(rdf) 
    assert not (rdfs_1 == rdfs_2)

    # test instance type
    assert not (rdfs_1 == rdf)

    # test sequence
    rdfs_2 = rdfs_1[1:]
    assert not (rdfs_1 == rdfs_2)

# WIP: Test on whole molecule and prmtop reading
def test_molecule():
    rdfs = rdfset_from_dir('tests/data/molecule/febpy_densities_00/', 
                           prmtop='tests/data/molecule/FEBPY_solv.prmtop')
    rdfs.get_dv(qvec=np.arange(0, 6, 0.1))

